from . import args
from .client import DuploClient
from .errors import DuploError, DuploFailedResource
from .commander import get_parser, Command
import math
import time

class DuploCommand():
  def __init__(self, duplo: DuploClient):
    self.duplo = duplo
  
  def __call__(self, *args):
    pass

class DuploResource():

  def __init__(self, duplo: DuploClient):
    self.duplo = duplo
    self.__logger = None
  
  def __call__(self, cmd: str, *args):
    command = self.command(cmd)
    parser = get_parser(command)
    parsed_args = parser.parse_args(args)
    return command(**vars(parsed_args))

  # something is off and the logs will duplicate if we do this  
  # @property
  # def logger(self):
  #   if not self.__logger:
  #     self.__logger = self.duplo.logger_for(self.__class__.__name__)
  #   return self.__logger
  
  def command(self, name: str):
    if not (command := getattr(self, name, None)):
      raise DuploError(f"Invalid command: {name}")
    return command
  
  def wait(self, wait_check: callable, timeout: int=None, poll: int=10):
    timeout = timeout or self.duplo.timeout
    exp = math.ceil(timeout / poll)
    for _ in range(exp):
      try:
        wait_check()
        break
      except DuploFailedResource as e:

        raise e
      except DuploError as e:
        if e.message:
          self.duplo.logger.info(e)
        time.sleep(poll)
      except KeyboardInterrupt as e:
        raise e
    else:
      raise DuploError("Timed out waiting", 404)
    
class DuploResourceV2(DuploResource):

  def name_from_body(self, body):
    return body["Name"]
  def endpoint(self, path: str=None):
    return path
  @Command()
  def list(self):
    """Retrieve a list of all services in a tenant."""
    response = self.duplo.get(self.endpoint(self.paths["list"]))
    return response.json()
  @Command()
  def find(self, 
           name: args.NAME):
    """Find a resource by name.
    
    Args:
      name (str): The name of the resource to find.
    Returns: 
      The resource object.
    Raises:
      DuploError: If the resource could not be found.
    """
    try:
      return [s for s in self.list() if self.name_from_body(s) == name][0]
    except IndexError:
      raise DuploError(f"{self.kind} '{name}' not found", 404)
      
  
class DuploTenantResourceV2(DuploResourceV2):
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.__tenant = None
    self.tenant_svc = duplo.load('tenant')
  @property
  def tenant(self):
    if not self.__tenant:
      self.__tenant = self.tenant_svc.find(self.duplo.tenant)
    return self.__tenant
  
  def endpoint(self, path: str=None):
    tenant_id = self.tenant["TenantId"]
    p = f"subscriptions/{tenant_id}/{path}"
    return p
  

class DuploTenantResourceV3(DuploResource):
  def __init__(self, duplo: DuploClient, slug: str):
    super().__init__(duplo)
    self.__tenant = None
    self.tenant_svc = duplo.load('tenant')
    self.slug = slug
    self.wait_timeout = 200
    self.wait_poll = 10
  @property
  def tenant(self):
    if not self.__tenant:
      self.__tenant = self.tenant_svc.find(self.duplo.tenant)
    return self.__tenant
  
  @Command()
  def list(self):
    """Retrieve a list of all resources in a tenant."""
    response = self.duplo.get(self.endpoint())
    return response.json()
  
  @Command()
  def find(self, 
           name: args.NAME):
    """Find a V3 resource by name.
    
    Args:
      name (str): The name of the resource to find.
    Returns: 
      The resource.
    Raises:
      DuploError: If the resource could not be found.
    """
    response = self.duplo.get(self.endpoint(name))
    return response.json()
  
  @Command()
  def delete(self, 
             name: args.NAME):
    """Delete a V3 resource by name.
    
    Args:
      name (str): The name of the resource to delete.
    Returns: 
      A success message.
    Raises:
      DuploError: If the resource could not be found or deleted. 
    """
    self.duplo.delete(self.endpoint(name))
    return {
      "message": f"{self.slug}/{name} deleted"
    }
  
  @Command()
  def create(self, 
             body: args.BODY,
             wait: args.WAIT=False,
             wait_check: callable=None):
    """Create a V3 resource by name.
    
    Args:
      body (str): The resource to create.
    Returns: 
      Success message.
    Raises:
      DuploError: If the resource could not be created.
    """
    name = self.name_from_body(body)
    response = self.duplo.post(self.endpoint(), body)
    if wait:
      self.wait(wait_check or (lambda: self.find(name)), self.wait_timeout, self.wait_poll)
    return response.json()
  
  @Command()
  def update(self, 
             body: args.BODY):
    """Update a V3 resource by name.
    
    Args:
      body (str): The resource to update.
    Returns: 
      Success message.
    Raises:
      DuploError: If the resource could not be created.
    """
    name = self.name_from_body(body)
    response = self.duplo.put(self.endpoint(name), body)
    return response.json()
  
  def name_from_body(self, body):
    return body["metadata"]["name"]

  def endpoint(self, name: str=None, path: str=None):
    tenant_id = self.tenant["TenantId"]
    p = f"v3/subscriptions/{tenant_id}/{self.slug}"
    if name:
      p += f"/{name}"
    if path:
      p += f"/{path}"
    return p



import time
from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError, DuploFailedResource
from duplocloud.commander import Command, Resource
from json import dumps, loads
import duplocloud.args as args

_STATUS_CODES = {
  "1": "Running",
  "3": "Pending",
  "4": "Waiting",
  "6": "Deleted",
  "7": "Failed",
  "11": "Succeeded"
}

@Resource("service")
class DuploService(DuploTenantResourceV2):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.paths = {
      "list": "GetReplicationControllers"
    }
    self.__pod_svc = self.duplo.load("pod")
    self.__old = None
    self.__updates = None
    self.__old_replicaset = None
    
  @Command()
  def update(self, 
             name: args.NAME,
             body: args.BODY = None,
             patches: args.PATCHES = None,
             wait: args.WAIT = False):
    """Update a service."""
    if (name is None and body is None):
      raise DuploError("No arguments provided to update service", 400)
    if name and body:
      body["Name"] = name
    if body is None:
      body = self.find(name)
    # before patching, save the original body if we need to wait
    old = None
    if wait:
      old = self.find(name) # should use the cache
      old["Replicaset"] = self.current_replicaset(name)
    if patches:
      body = self.duplo.jsonpatch(body, patches)
    if ((ttags := body["Template"].get("AllocationTags", None))
        and not body.get("AllocationTags", None)):
      body["AllocationTags"] = ttags
    self.duplo.post(self.endpoint("ReplicationControllerChangeAll"), body)
    if wait:
      self.wait(old, body)
    return {
      "message": f"Successfully updated service '{body['Name']}'"
    }
  
  @Command()
  def create(self,
             body: args.BODY,
             wait: args.WAIT = False):
    """Create a service."""
    self.duplo.post(self.endpoint("ReplicationControllerUpdate"), body)
    if wait:
      self.wait(lambda: self.find(body["Name"]))
    return {
      "message": f"Successfully created service '{body['Name']}'"
    }

  @Command()
  def delete(self,
             name: args.NAME):
    """Delete a service."""
    body = {
      "Name": name,
      "State": "delete"
    }
    self.duplo.post(self.endpoint("ReplicationControllerUpdate"), body)
    return {
      "message": f"Successfully deleted service '{name}'"
    }

  @Command()
  def update_replicas(self, 
                      name: args.NAME,
                      replica: args.REPLICAS,
                      wait: args.WAIT = False):
    """Update number of replicas for a service.

    Args:
        name (str): The name of the service to update.
        replica (str): Number of replicas to set for service.
    """
    service = self.find(name)
    data = {
      "Name": name,
      "Replicas": replica,
      "AllocationTags": service["Template"].get("AllocationTags", "")
    }
    self.duplo.post(self.endpoint("ReplicationControllerChange"), data)
    if wait:
      self.wait(service, data)
    return {"message": f"Successfully updated replicas for service '{name}'"} 
  
  @Command()
  def update_image(self, 
                   name: args.NAME, 
                   image: args.IMAGE,
                   wait: args.WAIT = False):
    """Update the image of a service.
    
    Args:
      name (str): The name of the service to update.
      image (str): The new image to use for the service.
    """
    service = self.find(name)
    current_image =  self.image_from_body(service)

    # needed before update starts, not needed if not waiting
    if wait:
      service["Replicaset"] = self.current_replicaset(name)

    if(current_image == image):
      self.duplo.post(self.endpoint(f"ReplicationControllerReboot/{name}"))
    else:
      data = {
        "Name": name,
        "Image": image,
        "AllocationTags": service["Template"].get("AllocationTags", "")
      }
      self.duplo.post(self.endpoint("ReplicationControllerChange"), data)
    if wait:
      self.wait(service, data)
    return {"message": f"Successfully updated image for service '{name}'"}
 
  @Command()
  def update_env(self, 
                 name: args.NAME,
                 setvar: args.SETVAR, 
                 strategy: args.STRATEGY,
                 deletevar: args.DELETEVAR,
                 wait: args.WAIT = False):
    """Update the environment variables of a service.

    Args:
      name (str): The name of the service to update.
      setvar/-V (list): A list of key value pairs to set as environment variables.
      strategy/strat (str): The merge strategy to use for env vars. Valid options are "merge" or "replace".  Default is merge.
      deletevar/-D (list): A list of keys to delete from the environment variables.
    """
    service = self.find(name)
    service["Replicaset"] = self.current_replicaset(name)
    currentDockerconfig = loads(service["Template"]["OtherDockerConfig"])
    currentEnv = currentDockerconfig.get("Env", [])
    newEnv = []
    if setvar is not None:
      newEnv = [{"Name": i[0], "Value": i[1]} for i in setvar]
    if strategy == 'merge':
      d = {d['Name']: d for d in currentEnv + newEnv}
      mergedvars = list(d.values())
      currentDockerconfig['Env'] = mergedvars
    else:
      currentDockerconfig['Env'] = newEnv
    if deletevar is not None:
      for key in deletevar:
        currentDockerconfig['Env'] = [d for d in currentDockerconfig['Env'] if d['Name'] != key]
    payload = {
      "Name": name,
      "OtherDockerConfig": dumps(currentDockerconfig),
      "allocationTags": service["Template"].get("AllocationTags", "")
    }
    self.duplo.post(self.endpoint("ReplicationControllerChange"), payload)
    if wait:
      self.wait(service, payload)
    return {"message": "Successfully updated environment variables for services"}
    
  
  @Command()
  def bulk_update_image(self, 
                  serviceimage: args.SERVICEIMAGE):
    """Update multiple services.
    
    Args:
      serviceimage/-S (string): takes n sets of two arguments, service name and image name. e.g -S service1 image1:tag -S service2 image2:tag
    """
    payload = []
    for i in serviceimage:
      servicepair = dict([i])
      for name, image in servicepair.items():
        payloaditem = {}
        service = self.find(name)
        allocation_tags = service["Template"]["AllocationTags"]
        payloaditem["Name"] = name
        payloaditem["Image"] = image
        payloaditem["AllocationTags"] = allocation_tags
        payload.append(payloaditem)
    self.duplo.post(self.endpoint("ReplicationControllerBulkChangeAll"), payload)
    return {"message": "Successfully updated images for services"}

  @Command()
  def restart(self, 
              name: args.NAME):
    """Restart a service.
    
    Args:
      name (str): The name of the service to restart.
    Returns: 
      A success message if the service was restarted successfully.
    Raises:
      DuploError: If the service could not be restarted.
    """
    self.duplo.post(self.endpoint(f"ReplicationControllerReboot/{name}"))
    return {"message": f"Successfully restarted service '{name}'"}
  
  @Command()
  def stop(self, 
           name: args.NAME):
    """Stop a service.
    
    Args:
      name (str): The name of the service to stop.
    Returns: 
      A success message if the service was stopped successfully.
    Raises:
      DuploError: If the service could not be stopped.
    """
    self.duplo.post(self.endpoint(f"ReplicationControllerStop/{name}"))
    return {"message": f"Successfully stopped service '{name}'"}
  
  @Command()
  def start(self, 
            name: args.NAME):
    """Start a service.
    
    Args:
      name (str): The name of the service to start.
    Returns: 
      A success message if the service was started successfully.
    Raises:
      DuploError: If the service could not be started.
    """
    self.duplo.post(self.endpoint(f"ReplicationControllerstart/{name}"))
    return {"message": f"Successfully started service '{name}'"}

  @Command()
  def pods(self, 
           name: args.NAME):
    """Get the pods for a service.
    
    Args:
      name (str): The name of the service to get pods for.
    Returns: 
      A list of pods for the service.
    Raises:
      DuploError: If the service could not be found.
    """
    pods = self.__pod_svc.list()
    return [
      pod for pod in pods
      if pod["Name"] == name and pod["ControlledBy"]["QualifiedType"] == "kubernetes:apps/v1/ReplicaSet"
    ]
  
  @Command()
  def logs(self,
           name: args.NAME,
           wait: args.WAIT = False):
    """Get the logs for a service."""
    def show_logs():
      pods = self.pods(name)
      for pod in pods:
        self.__pod_svc.logs(pod=pod)
    if wait:
      try:
        while True:
          show_logs()
          time.sleep(3)
      except KeyboardInterrupt:
        pass
    else:
      show_logs()

  def current_replicaset(self, name: str):
    """Get the current replicaset for a service.
    
    Args:
      name (str): The name of the service to get replicaset for.
    Returns: 
      The current replicaset for the service.
    Raises:
      DuploError: If the service could not be found.
    """
    pods = self.pods(name)
    return pods[0]["ControlledBy"]["NativeId"]
  
  def image_from_body(self, body):
    """Get the image from a service body.
    
    Args:
      body (dict): The body of the service.
    Returns: 
      The image for the service.
    """
    tpl = body.get("Template", {})
    containers = tpl.get("Containers", [])
    for c in containers:
      if c["Name"] == body["Name"]:
        return c["Image"]
    else:
      return body.get("DockerImage", body.get("Image", None))

  def wait(self, old, updates):
    """Wait for a service to update."""
    name = old["Name"]
    new_img = self.image_from_body(updates)
    old_img = self.image_from_body(old)
    new_replicas = updates.get("Replicas", None)
    replicas_changed = (old["Replicas"] != new_replicas) if new_replicas else False
    image_changed = (old_img != new_img) if new_img else False
    new_conf = updates.get("OtherDockerConfig", None)
    conf_changed = (old["Template"].get("OtherDockerConfig", None) != new_conf) if new_conf else False
    rollover = False
    if image_changed or conf_changed:
      rollover = True
    def wait_check():
      svc = self.find(name)
      replicas = svc["Replicas"]
      # make sure the change has been applied
      if (image_changed and self.image_from_body(svc) != new_img):
        raise DuploError(f"Service {name} waiting for image update", 400)
      if (replicas_changed and replicas != new_replicas):
        raise DuploError(f"Service {name} waiting for replicas update", 400)
      if (conf_changed and svc["Template"].get("OtherDockerConfig", None) != new_conf):
        raise DuploError(f"Service {name} waiting for pod to update", 400)
      pods = self.pods(name)
      faults = self.tenant_svc.faults()
      running = 0
      for p in pods:
        if p["ControlledBy"]["NativeId"] == old.get("Replicaset", None) and rollover:
          continue
        for f in faults:
          if f["Resource"]["Name"] == p["InstanceId"]:
            raise DuploFailedResource(f"Service {name} raised a fault.\n{f['Description']}")
        if ((p["CurrentStatus"] == p["DesiredStatus"]) and p["DesiredStatus"] == 1):
          running += 1
      if replicas != running:
        raise DuploError(f"Service {name} waiting for pods {running}/{replicas}", 400)
    super().wait(wait_check, 400, 11)



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Resource

@Resource("configmap")
class DuploConfigMap(DuploTenantResourceV3):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "k8s/configmap")




#!/usr/bin/env python3
import datetime
import argparse
import sys
import os
HERE = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(HERE))
from gha import GithubRepo

UNRELEASED = "## [Unreleased]"
CHANGELOG = 'CHANGELOG.md'

def get_changelog():
  with open(CHANGELOG, 'r') as f:
    return f.read()
  
def release_notes(changelog):
  inblock = False
  msg = []
  for line in changelog.splitlines():
    if line.startswith(UNRELEASED):
      inblock = True
      continue
    if inblock and line.startswith("## ["):
      break
    if inblock:
      msg.append(line.strip())
  return "\n".join(msg)

def replace_unreleased(changelog, version):
  # get todays date formatted like 2023-03-05
  date = datetime.datetime.now().strftime("%Y-%m-%d")
  new_notes = f"{UNRELEASED}\n\n## [{version}] - {date}"
  return changelog.replace(UNRELEASED, new_notes)

def save_github_output(notes, version, tag):
  """Save the tag and version in github output file"""
  outputs = os.environ.get('GITHUB_OUTPUT', './.github/output')
  with open(outputs, 'a') as f:
    f.write(f"version={version}\n")
    f.write(f"tag={tag}\n")

def publish(action, push, token):
  repo = GithubRepo(token)
  v = repo.bump_version(action)
  t = f"v{v}"
  c = get_changelog()
  n = release_notes(c)
  c = replace_unreleased(c, v)
  if push != "true":
    v = repo.ref
    t = v
    repo.dist_file(CHANGELOG, c)
  elif push == "true":
    print(f"Pushing changes for v{v}")
    repo.publish(t, CHANGELOG, c)    
  save_github_output(n, v, t)
  repo.dist_file('notes.md', n)
  
if __name__ == '__main__':
  parser = argparse.ArgumentParser(
    prog='version-bumper',
    description='A duploctl version bumper.',
  )
  parser.add_argument('--action', type=str, help='The type of version bump to perform.', default="patch")
  parser.add_argument('--push', type=str, help='Push to remote?', default="false")
  parser.add_argument('--token', type=str, help='Github token', default=os.environ.get('GITHUB_TOKEN'))
  args = parser.parse_args()
  publish(**vars(args))



import pytest
from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from .conftest import get_test_data

class TestCronjobs:
  
  @pytest.mark.integration
  @pytest.mark.order(7)
  @pytest.mark.dependency(
    name="update_cronjob_image", 
    depends=["find_tenant_resource"],
    scope='session')
  def test_update_image(self, duplo: DuploClient):
    kind = "cronjob"
    r = duplo.load(kind)
    body = get_test_data(kind)
    name = r.name_from_body(body)
    try:
      r.update_image(name, "ubuntu:latest")
      o = r.find(name)
      assert o["spec"]["jobTemplate"]["spec"]["template"]["spec"]["containers"][0]["image"] == "ubuntu:latest"
    except DuploError as e:
      pytest.fail(f"Failed to find tenant {name}: {e}")
  
  @pytest.mark.integration
  @pytest.mark.order(7)
  @pytest.mark.dependency(
    name="update_cronjob_schedule", 
    depends=["find_tenant_resource"], 
    scope='session')
  def test_update_schedule(self, duplo: DuploClient):
    kind = "cronjob"
    r = duplo.load(kind)
    body = get_test_data(kind)
    name = r.name_from_body(body)
    try:
      r.update_schedule(name, "1 1 * * 0")
      o = r.find(name)
      assert o["spec"]["schedule"] == "1 1 * * 0"
    except DuploError as e:
      pytest.fail(f"Failed to find tenant {name}: {e}")
  
  




from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("cronjob")
class DuploCronJob(DuploTenantResourceV3):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "k8s/cronJob")

  @Command()
  def update_image(self, 
                   name: args.NAME, 
                   image: args.IMAGE):
    """Update the image of a cronjob.
    
    Args:
      name (str): The name of the cronjob to update.
      image (str): The new image to use for the cronjob.
    """
    cronjob = self.find(name)
    cronjob["spec"]["jobTemplate"]["spec"]["template"]["spec"]["containers"][0]["image"] = image
    self.update(cronjob)
    return {"message": f"Successfully updated image for cronjob '{name}'"}
  
  @Command()
  def update_schedule(self, 
                      name: args.NAME, 
                      cronschedule: args.CRONSCHEDULE):
    """Update the schedule of a cronjob.
    
    Args:
      name (str): The name of the cronjob to update.
      schedule (str): The new schedule to use for the cronjob.
    """
    cronjob = self.find(name)
    cronjob["spec"]["schedule"] = cronschedule
    self.update(cronjob)
    return {"message": f"Successfully updated cron-schedule for cronjob '{name}'"}



from typing import Any
import requests
from git import Repo
import os
from packaging.version import Version
import semver

CWD = os.getcwd()

class GithubRepo:
  def __init__(self, token, repo_name="duploctl"):
    self.repo = Repo(CWD)
    self.url = f"https://api.github.com/repos/duplocloud/{repo_name}/git"
    self.headers = {
      "Accept": "application/vnd.github.v3+json",
      "Authorization": f"token {token}",
    }

  @property
  def ref(self):
    return self.repo.head.reference

  def publish(self, tag, file, content) -> Any:
    base_tree = self.repo.head.object.hexsha
    tree = self.create_tree(base_tree, file, content)
    print(tree)
    commit = self.create_commit(base_tree, tree, f"Bump version to {tag}")
    print(commit)
    ref = self.update_main(commit)
    print(ref)
    tag = self.create_tag(tag, commit)
    print(tag)

  def create_tree(self, base_tree, file, content):
    r = requests.post(f"{self.url}/trees", headers=self.headers, json={
    "base_tree": str(base_tree),
    "tree": [{
      "path": file,
      "mode": "100644",
      "type": "blob",
      "content": content
    }]})
    return r.json()

  def create_commit(self, base_tree, tree, message):
    r = requests.post(f"{self.url}/commits", headers=self.headers, json={
      "message": message,
      "tree": tree["sha"],
      "parents": [base_tree],
    })
    return r.json()
  
  def update_main(self, commit):
    r = requests.patch(f"{self.url}/refs/heads/main", headers=self.headers, json={
      "sha": commit["sha"]
    })
    return r.json()
  
  def create_tag(self, tag, commit):
    r = requests.post(f"{self.url}/refs", headers=self.headers, json={
      "ref": f"refs/tags/{tag}",
      "sha": commit["sha"]
    })
    ref = r.json()

  def latest_tag(self):
    latest = Version("0.0.0")
    for t in self.repo.tags:
      v = Version(t.name[1:])
      if v > latest:
        latest = v
    return semver.VersionInfo.parse(str(latest))
  
  def bump_version(self, action: str="patch"):
    latest = self.latest_tag()
    if action == "major":
      return latest.bump_major()
    if action == "minor":
      return latest.bump_minor()
    if action == "patch":
      return latest.bump_patch()
    return latest
  
  def dist_file(self, file, content):
    os.makedirs('dist', exist_ok=True)
    with open(f"dist/{file}", 'w') as f:
      f.write(content)



from duplocloud.client import DuploClient
from duplocloud.errors import DuploError, DuploExpiredCache
from duplocloud.resource import DuploResource
from duplocloud.commander import Command, Resource
import duplocloud.args as args
import os
from pathlib import Path
import yaml
import configparser
import webbrowser
from datetime import datetime
import jwt

@Resource("jit")
class DuploJit(DuploResource):
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)

  @Command()
  def aws(self, nocache: bool = None):
    """Retrieve aws session credentials for current user."""
    sts = None
    path = None
    k = self.duplo.cache_key_for("aws-creds")
    nc = nocache if nocache is not None else self.duplo.nocache

    # check if admin or choose tenant
    if self.duplo.isadmin:
      path = "adminproxy/GetJITAwsConsoleAccessUrl"
    else:
      t = self.duplo.load("tenant")
      tenant = t.find(self.duplo.tenant)
      path = f"subscriptions/{tenant['TenantId']}/GetAwsConsoleTokenUrl"

    # try and get those creds
    try:
      if nc:
        sts = self.duplo.get(path).json()
      else:
        sts = self.duplo.get_cached_item(k)
        if self.duplo.expired(sts.get("Expiration", None)):
          raise DuploExpiredCache(k)
    except DuploExpiredCache:
      sts = self.duplo.get(path).json()
      if "Expiration" not in sts:
        sts["Expiration"] = self.duplo.expiration()
      self.duplo.set_cached_item(k, sts)
    sts["Version"] = 1
    return sts

  @Command()
  def k8s(self,
          planId: args.PLAN = None):
    """Retrieve k8s session credentials for current user."""
    # either plan or tenant in cache key
    pt = planId if planId else self.duplo.tenant
    k = self.duplo.cache_key_for(f"plan,{pt},k8s-creds")
    creds = None
    try:
      if self.duplo.nocache:
        ctx = self.k8s_context(planId)
        creds = self.__k8s_exec_credential(ctx)
      else:
        creds = self.duplo.get_cached_item(k)
        exp = creds.get("status", {}).get("expirationTimestamp", None)
        if self.duplo.expired(exp):
          raise DuploExpiredCache(k)
    except DuploExpiredCache:
      ctx = self.k8s_context(planId)
      creds = self.__k8s_exec_credential(ctx)
      self.duplo.set_cached_item(k, creds)
    return creds
  
  @Command()
  def update_kubeconfig(self,
                        planId: args.PLAN = None):
    """Update kubeconfig"""
    # first get the kubeconfig file and parse it
    kubeconfig_path = os.environ.get("KUBECONFIG", f"{Path.home()}/.kube/config")
    kubeconfig = (yaml.safe_load(open(kubeconfig_path, "r")) 
                  if os.path.exists(kubeconfig_path) 
                  else self.__empty_kubeconfig())
    # load the cluster config info
    ctx = self.k8s_context(planId)
    if self.duplo.isadmin:
      ctx["Name"] = ctx["Name"].removeprefix("duploinfra-")
      ctx["cmd_arg"] = "--plan"
    else:
      ctx["Name"] = self.duplo.tenant
      ctx["cmd_arg"] = "--tenant"
    # add the cluster, user, and context to the kubeconfig
    self.__add_to_kubeconfig("clusters", self.__cluster_config(ctx), kubeconfig)
    self.__add_to_kubeconfig("users", self.__user_config(ctx), kubeconfig)
    self.__add_to_kubeconfig("contexts", self.__context_config(ctx), kubeconfig)
    kubeconfig["current-context"] = ctx["Name"]
    # write the kubeconfig back to the file
    with open(kubeconfig_path, "w") as f:
      yaml.safe_dump(kubeconfig, f)
    return {"message": f"kubeconfig updated successfully to {kubeconfig_path}"}
  
  @Command()
  def update_aws_config(self,
                        name: args.NAME):
    """Update aws config"""
    config = os.environ.get("AWS_CONFIG_FILE", f"{Path.home()}/.aws/config")
    cp = configparser.ConfigParser()
    cp.read(config)
    prf = f'profile {name}'
    msg = f"aws profile named {name} already exists in {config}"
    try:
      cp[prf]
    except KeyError:
      cmd = self.duplo.build_command("duploctl", "jit", "aws")
      cp.add_section(prf)
      cp.set(prf, 'region', os.getenv("AWS_DEFAULT_REGION", "us-west-2"))
      cp.set(prf, 'credential_process', " ".join(cmd))
      with open(config, 'w') as configfile:
        cp.write(configfile)
      msg = f"aws profile named {name} was successfully added to {config}"
    return {"message": msg}

  @Command()
  def web(self):
    b = self.duplo.browser
    wb = webbrowser if not b else webbrowser.get(b)
    sts = self.aws(nocache=True)
    wb.open(sts["ConsoleUrl"], new=0, autoraise=True)
    return {
      "message": "Opening AWS console in browser"
    }
  
  @Command()
  def k8s_context(self, 
                  planId: args.PLAN = None):
    """Get k8s context
    
    Gets context based on planId or tenant name or admin or nonadmin. 

    Args:
      planId (str): The planId of the infrastructure.
    
    Returns:
      dict: The k8s context.
    """
    tenant = None
    tenant_id = None
    tenant_name = self.duplo.tenant

    # don't even like try sometimes
    if not self.duplo.isadmin and not tenant_name:
      raise DuploError("--tenant is required", 300)
    if planId is None and not tenant_name:
      raise DuploError("--plan or --tenant is required", 300)
    
    # if we need to discover the planId or you are not admin, then we get the tenant
    if (planId is None and tenant_name) or not self.duplo.isadmin:
      tenant_svc = self.duplo.load("tenant")
      tenant = tenant_svc.find(tenant_name)
      planId = tenant.get("PlanID") if not planId else planId
      tenant_id = tenant.get("TenantId")

    # choose the correct path
    path = (f"v3/admin/plans/{planId}/k8sConfig"
            if self.duplo.isadmin
            else f"/v3/subscriptions/{tenant_id}/k8s/jitAccess")
    response = self.duplo.get(path)
    return response.json()

  def __k8s_exec_credential(self, ctx):
    cluster = {
      "server": ctx["ApiServer"],
      "config": None
    }
    if ctx["K8Provider"] == 0 and (ca := ctx.get("CertificateAuthorityDataBase64", None)):
      cluster["certificate-authority-data"] = ca

    t = jwt.decode(jwt=ctx["Token"],algorithms=["HS256"],options={"verify_signature": False})
    exp = datetime.fromtimestamp(t["exp"]).strftime('%Y-%m-%dT%H:%M:%S+00:00')

    return {
      "kind": "ExecCredential",
      "apiVersion": "client.authentication.k8s.io/v1beta1",
      "spec": {
        "cluster": cluster,
        "interactive": False
      },
      "status": {
        "token": ctx["Token"],
        "expirationTimestamp": exp
      }
    }

  def __cluster_config(self, ctx):
    """Build a kubeconfig cluster object"""
    cluster = {
      "server": ctx["ApiServer"]
    }
    if ctx["K8Provider"] == 0 and (ca := ctx.get("CertificateAuthorityDataBase64", None)):
      cluster["certificate-authority-data"] = ca
    elif ctx["K8Provider"] == 1:
      cluster["insecure-skip-tls-verify"] = True
    return {
      "name": ctx["Name"],
      "cluster": cluster
    }
  
  def __user_config(self, ctx):
    """Build a kubeconfig user object"""
    cmd = self.duplo.build_command("jit", "k8s", ctx["cmd_arg"], ctx["Name"])
    return {
      "name": ctx["Name"],
      "user": {
        "exec": {
          "apiVersion": "client.authentication.k8s.io/v1beta1",
          "installHint": """
Install duploctl for use with kubectl by following
https://github.com/duplocloud/duploctl
""",
          "command": "duploctl",
          "args": cmd
        }
      }
    }
  
  def __context_config(self, ctx):
    """Build a kubeconfig context object"""
    return {
      "name": ctx["Name"],
      "context": {
        "cluster": ctx["Name"],
        "user": ctx["Name"],
        "namespace": ctx["DefaultNamespace"]
      }
    }
  
  def __add_to_kubeconfig(self, section, item, kubeconfig):
    """Add an item to a kubeconfig section if it is not already present"""
    if item not in kubeconfig[section]:
      kubeconfig[section].append(item)

  def __empty_kubeconfig(self):
    return {
      "apiVersion": "v1",
      "kind": "Config",
      "preferences": {},
      "clusters": [],
      "users": [],
      "contexts": []
    }
  

    



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("lambda")
class DuploLambda(DuploTenantResourceV2):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
  
  @Command()
  def list(self):
    """Retrieve a list of all lambdas in a tenant."""
    tenant_id = self.tenant["TenantId"]
    tenant_name = self.tenant["AccountName"]
    response = self.duplo.get(f"subscriptions/{tenant_id}/GetLambdaFunctions")
    if (data := response.json()):
      return data
    else:
      raise DuploError(f"No lambda functions found in tenant '{tenant_name}'", 404)
  
  @Command()
  def find(self, 
           name: args.NAME):
    """Find a Lambda function by name.
    
    Args:
      name (str): The name of the lambda to find.
    Returns: 
      The lambda object.
    Raises:
      DuploError: If the lambda could not be found.
    """
    try:
      return [s for s in self.list() if s["FunctionName"] == name][0]
    except IndexError:
      raise DuploError(f"Lambda '{name}' not found", 404)
    
  @Command()
  def create(self, 
             body: args.BODY,
             wait: args.WAIT = False):
    """Create a new tenant."""
    def wait_check():
      name = self.name_from_body(body)
      self.find(name)
    tenant_id = self.tenant["TenantId"]
    self.duplo.post(f"subscriptions/{tenant_id}/CreateLambdaFunction", body)
    if wait:
      self.wait(wait_check, 400)
    return {
      "message": f"Lambda {body['FunctionName']} created"
    }
  
  @Command()
  def delete(self, 
             name: args.NAME):
    """Delete a lambda."""
    tenant_id = self.tenant["TenantId"]
    self.duplo.post(f"subscriptions/{tenant_id}/DeleteLambdaFunction/{name}")
    return {
      "message": f"Lambda {name} deleted"
    }

  @Command()
  def update_image(self, 
                   name: args.NAME, 
                   image: args.IMAGE):
    """Update the image of a lambda.
    
    Args:
      name (str): The name of the lambda to update.
      image (str): The new image to use for the lambda.
    """
    tenant_id = self.tenant["TenantId"]
    data = {
      "FunctionName": name,
      "ImageUri": image
    }
    response = self.duplo.post(f"subscriptions/{tenant_id}/UpdateLambdaFunction", data)
    return response.json()

  @Command()
  def update_s3(self, 
                   name: args.NAME, 
                   bucket: args.S3BUCKET,
                   key: args.S3KEY):
    """Update the s3 bucket and key of a lambda.
    
    Args:
      name (str): The name of the lambda to update.
      s3bucket (str): The s3 bucket to use for the lambda.
      s3key (str): The s3 key (file path) to use for the lambda.
    """
    tenant_id = self.tenant["TenantId"]
    data = {
      "FunctionName": name,
      "S3Bucket": bucket,
      "S3Key": key
    }
    response = self.duplo.post(f"subscriptions/{tenant_id}/UpdateLambdaFunction", data)
    return response.json()

  def name_from_body(self, body):
    prefix = f"duploservices-{self.duplo.tenant}"
    name =  body["FunctionName"]
    if not name.startswith(prefix):
      name = f"{prefix}-{name}"
    return name



#!/usr/bin/env python3
import argparse
import sys
import os
HERE = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(HERE))
import requests
import toml
import re
from importlib.metadata import version
from gha import GithubRepo

class HomebrewFormula:

  def __init__(self, token, tag=None):
    self.repo = GithubRepo(token, "homebrew-tap")
    f = open('pyproject.toml')
    self.pyproject = toml.load(f)
    self.repo_url = self.pyproject['project']['urls']['Repository']
    self.description = self.pyproject['project']['description']
    self.tpl_file = 'scripts/formula.tpl.rb'
    self.out_file = 'duploctl.rb'
    self.version = tag.replace("v", "") if tag else self.repo.latest_tag()

  def publish(self, push="false"):
    formula = self.build_formula()
    print(formula)
    if push == "true":
      print("Pushing to remote")
      self.repo.publish(f"duploctl-v{self.version}", 'Formula/duploctl.rb', formula)
    else:
      self.repo.dist_file(self.out_file, formula)

  def make_resource(self, name):
    v = version(name)
    url = f"https://pypi.org/pypi/{name}/{v}/json"
    response = requests.get(url)
    data = response.json()
    sdist = list(filter(lambda x: x['packagetype'] == 'sdist', data['urls']))[0]
    return f"""
      resource "{name}" do
        url "{sdist['url']}"
        sha256 "{sdist['digests']['sha256']}"
      end
    """
  
  def make_resources(self):
    pattern = '|'.join(map(re.escape, ['>=', '<=', '==', '!=']))
    resources = []
    for dep in self.pyproject['project']['dependencies']:
      name, _ = re.split(pattern, dep, 1)
      res = self.make_resource(name)
      resources.append(res)
    return "".join(resources)
  
  def get_shas(self):
    url = f"{self.repo_url}/releases/download/v{self.version}/checksums.txt"
    response = requests.get(url)
    checksums = response.text.splitlines()
    linux_sha = None 
    macos_sha = None
    pip_sha = None
    for line in checksums:
      l = line.split(" ")
      if 'darwin' in line:
        macos_sha = l[0]
      elif 'linux' in line:
        linux_sha = l[0]
      elif 'duplocloud_client' in line:
        pip_sha = l[0]
    return (linux_sha, macos_sha, pip_sha)
  
  def build_formula(self):
    with open(self.tpl_file, 'r') as tpl_file:
      tpl = tpl_file.read()
      linux_sha, macos_sha, pip_sha = self.get_shas()
      resources = self.make_resources()
      formula = tpl.format(
        repo_url=self.repo_url,
        description=self.description,
        version=self.version, 
        linux_sha=linux_sha, 
        macos_sha=macos_sha,
        pip_sha=pip_sha,
        resources=resources
      )
      return formula
  
  
if __name__ == '__main__':
  parser = argparse.ArgumentParser(
    prog='tap-formula',
    description='Pushes formula to tap',
  )

  parser.add_argument('--tag', type=str, help='A version tag to use.', default=None)
  parser.add_argument('--push', type=str, help='Push to remote?', default="false")
  parser.add_argument('--token', type=str, help='Github token', default=os.environ.get('GITHUB_TOKEN'))

  args = parser.parse_args()

  print(args)

  brew = HomebrewFormula(args.token, args.tag)
  brew.publish(args.push)




import pytest
from duplocloud.errors import DuploError
import time

resources = [
  "hosts",
  "asg",
  "cronjob", 
  "job",
  "secret",
  "configmap",
  "lambda",
  "rds",
  "rds::rds-read"
]

@pytest.mark.parametrize("test_data", resources, indirect=True)
class TestTenantResources:

  @pytest.mark.integration
  @pytest.mark.order(5)
  @pytest.mark.dependency(
    name="create_tenant_resource", 
    scope='session')
  def test_creating_resource(self, test_data, duplo):
    (kind, data) = test_data
    r = duplo.load(kind)
    name = r.name_from_body(data)
    start_time = time.time()
    print(f"Creating {kind} '{name}' in '{duplo.tenant}'")
    # For some reason you'll get a 409 a bunch of times until the tenant is actually ready.
    while True:
      try:
        r.create(data, wait=True)
        print(f"{kind} '{name}' created in '{duplo.tenant}'")
        break
      except DuploError as e:
        elapsed_time = time.time() - start_time
        if elapsed_time > 45:
          pytest.fail(f"Failed to create {kind} after x seconds: {e}")
        else:
          print(f"Attempt failed: {e}. Retrying in 5 seconds...")
          time.sleep(5)

  @pytest.mark.integration
  @pytest.mark.order(6)
  @pytest.mark.dependency(
    name="find_tenant_resource", 
    depends=["create_tenant_resource"], 
    scope='session')
  def test_find_resource(self, duplo, test_data):
    (kind, data) = test_data
    r = duplo.load(kind)
    name = r.name_from_body(data)
    try:
      o = r.find(name)
      assert r.name_from_body(o) == name
      print(f"Found {kind} '{name}'")
    except DuploError as e:
      pytest.fail(f"Failed to find {kind} {name}: {e}")
  
  @pytest.mark.integration
  @pytest.mark.k8s
  @pytest.mark.order(8)
  @pytest.mark.dependency(
    name="delete_tenant_resource", 
    depends=["create_tenant_resource"], 
    scope='session')
  def test_delete_resource(self, duplo, test_data):
    (kind, data) = test_data
    r = duplo.load(kind)
    name = r.name_from_body(data)
    try:
      r.delete(name)
    except DuploError as e:
      pytest.fail(f"Failed to delete {kind}: {e}")




from duplocloud.client import DuploClient  # Importing necessary modules
from duplocloud.errors import DuploError, DuploFailedResource
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("job")  # Decorator to define a resource
class DuploJob(DuploTenantResourceV3):
  
  def __init__(self, duplo: DuploClient):  # Constructor method
    super().__init__(duplo, "k8s/job")  
    self.wait_timeout = 1000
    self.wait_poll = 3
    self.__pod_svc = self.duplo.load("pod")

  @Command()  
  def create(self, 
             body: args.BODY, 
             wait: args.WAIT = False):
    """Create a job."""
    name = self.name_from_body(body)
    a = 0
    s = 0
    f = 0
    def wait_check():
      nonlocal a, s, f
      job = self.find(name)
      active = job["status"].get("active", 0)
      succeeded = job["status"].get("succeeded", 0)
      failed = job["status"].get("failed", 0)
      completions = job["spec"].get("completions", 1)
      limit = job["spec"].get("backoffLimit", 6)
      cond = job["status"].get("conditions", [])
      cpl = [c for c in cond if c["type"] == "Complete" and c["status"] == "True"]
      fail = [c for c in cond if c["type"] == "Failed" and c["status"] == "True"]
      if (a != active or s != succeeded or f != failed):
        a = active
        s = succeeded
        f = failed
        self.duplo.logger.info(f"Job {name}: active({active}/{completions}), succeeded({succeeded}/{completions}), failed({failed}/{limit})")
      # make sure we can get pods and logs first
      pods_exist = (active > 0 or succeeded > 0 or failed > 0)
      pods = self.pods(name)
      if pods_exist and len(pods) == 0:
        raise DuploError(None)
      # display the logs
      for pod in pods:
        self.__pod_svc.logs(pod=pod)
      # make sure we got all of the logs
      pod_count = active + succeeded + failed
      if len(pods) != pod_count:
        raise DuploError(None)
      if len(fail) > 0:
        raise DuploFailedResource(f"Job {name} failed with {fail[0]['reason']}: {fail[0]['message']}")
      if not len(cpl) > 0:
        raise DuploError(None)
    super().create(body, wait, wait_check)
    return {
      "message": f"Job {name} ran successfully."
    }


  @Command()
  def pods(self, 
           name: args.NAME):
    """Get the pods for a service.
    
    Args:
      name (str): The name of the service to get pods for.
    Returns: 
      A list of pods for the service.
    Raises:
      DuploError: If the service could not be found.
    """
    pods = self.__pod_svc.list()
    return [
      pod for pod in pods
      if pod["Name"] == name and pod["ControlledBy"]["QualifiedType"] == "kubernetes:batch/v1/Job"
    ]
